﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace ContainerShip
{
    internal abstract class AbstractMap : IEquatable<AbstractMap>
    {
        private IDrawingObject _drawingObject = null;
        protected int[,] _map = null;
        protected int _width;
        protected int _height;
        protected float _size_x;
        protected float _size_y;
        protected readonly Random _random = new();
        protected readonly int _freeRoad = 0;
        protected readonly int _barrier = 1;
        public Bitmap CreateMap(int width, int height, IDrawingObject drawingObject)
        {
            _width = width;
            _height = height;
            _drawingObject = drawingObject;
            GenerateMap();
            while (!SetObjectOnMap())
            {
                GenerateMap();
            }
            return DrawMapWithObject();
        }
        public Bitmap MoveObject(Direction direction)
        {
            if (_drawingObject == null) return DrawMapWithObject();
            bool canMove = true;
            switch (direction)
            {
                case Direction.Left:
                    if (!CheckForBarriers(0, -1 * _drawingObject.Step, -1 * _drawingObject.Step, 0)) canMove = false;
                    break;
                case Direction.Right:
                    if (!CheckForBarriers(0, _drawingObject.Step, _drawingObject.Step, 0)) canMove = false;
                    break;
                case Direction.Up:
                    if (!CheckForBarriers(-1 * _drawingObject.Step, 0, 0, -1 * _drawingObject.Step)) canMove = false;
                    break;
                case Direction.Down:
                    if (!CheckForBarriers(_drawingObject.Step, 0, 0, _drawingObject.Step)) canMove = false;
                    break;
            }
            if (canMove)
            {
                _drawingObject.MoveObject(direction);
            }
            return DrawMapWithObject();
        }
        //Вспомогательная функция проверки
        private bool CheckForBarriers(float topBorder, float rightBorder, float leftBorder, float bottomBorder)
        {
            int top = Convert.ToInt32((_drawingObject.GetCurrentPosition().Top + topBorder)/_size_y);
            int bottom = Convert.ToInt32((_drawingObject.GetCurrentPosition().Bottom + bottomBorder)/_size_y);
            int right = Convert.ToInt32((_drawingObject.GetCurrentPosition().Right + rightBorder)/_size_x);
            int left = Convert.ToInt32((_drawingObject.GetCurrentPosition().Left + leftBorder)/_size_x);

            if (left < 0 || top < 0 || right >= _map.GetLength(0) || bottom >= _map.GetLength(1))
            {
                return false;
            }

            for (int i = top; i <= bottom; i++)
            {
                for (int j = left; j <= right; j++)
                {
                    if (_map[j, i] == 1) return false;
                }
            }
            return true;
        }
        private bool SetObjectOnMap()
        {
            if (_drawingObject == null || _map == null)
            {
                return false;
            }
            int x = _random.Next(0, 10);
            int y = _random.Next(0, 10);
            _drawingObject.SetObject(x, y, _width, _height);
            if (!CheckForBarriers(0, 0, 0, 0)) return false;
            return true;
        }
        private Bitmap DrawMapWithObject()
        {
            Bitmap bmp = new(_width, _height);
            if (_drawingObject == null || _map == null)
            {
                return bmp;
            }
            Graphics gr = Graphics.FromImage(bmp);
            for (int i = 0; i < _map.GetLength(0); ++i)
            {
                for (int j = 0; j < _map.GetLength(1); ++j)
                {
                    if (_map[i, j] == _freeRoad)
                    {
                        DrawRoadPart(gr, i, j);
                    }
                    else if (_map[i, j] == _barrier)
                    {
                        DrawBarrierPart(gr, i, j);
                    }
                }
            }
            _drawingObject.DrawingObject(gr);
            return bmp;
        }
        public bool Equals(AbstractMap? other)
        {
            if (_width != other._width)
            {
                return false;
            }
            if (_height != other._height)
            {
                return false;
            }
            if (_size_x != other._size_x)
            {
                return false;
            }
            if (_size_y != other._size_y)
            {
                return false;
            }
            for (int i = 0; i < _map.GetLength(0); i++)
            {
                for (int j = 0; j < _map.GetLength(1); j++)
                {
                    if (other._map[i, j] != _map[j, i]) return false;
                }
            }
            return true;
        }
        protected abstract void GenerateMap();
        protected abstract void DrawRoadPart(Graphics g, int i, int j);
        protected abstract void DrawBarrierPart(Graphics g, int i, int j);
    }
}
