﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace ContainerShip
{
    internal class MapWithSetShipsGeneric<T, U>
        where T : class, IDrawingObject, IEquatable<T>
       where U : AbstractMap
    {
        /// <summary>
        /// Ширина окна отрисовки
        /// </summary>
        private readonly int _pictureWidth;
        /// <summary>
        /// Высота окна отрисовки
        /// </summary>
        private readonly int _pictureHeight;
        /// <summary>
        /// Размер занимаемого объектом места (ширина)
        /// </summary>
        private readonly int _placeSizeWidth = 135;
        /// <summary>
        /// Размер занимаемого объектом места (высота)
        /// </summary>
        private readonly int _placeSizeHeight = 70;
        /// <summary>
        /// Набор объектов
        /// </summary>
        private readonly SetShipsGeneric<T> _setShips;
        /// <summary>
        /// Карта
        /// </summary>
        private readonly U _map;
        /// <summary>
        /// Конструктор
        /// </summary>
        /// <param name="picWidth"></param>
        /// <param name="picHeight"></param>
        /// <param name="map"></param>
        public MapWithSetShipsGeneric(int picWidth, int picHeight, U map)
        {
            int width = picWidth / _placeSizeWidth;
            int height = picHeight / _placeSizeHeight;
            _setShips = new SetShipsGeneric<T>(width * height);
            _pictureWidth = picWidth;
            _pictureHeight = picHeight;
            _map = map;
        }
        /// <summary>
        /// Вывод всего набора объектов
        /// </summary>
        /// <returns></returns>
        public Bitmap ShowSet()
        {
            Bitmap bmp = new(_pictureWidth, _pictureHeight);
            Graphics gr = Graphics.FromImage(bmp);
            DrawBackground(gr);
            DrawShips(gr);
            return bmp;
        }
        /// <summary>
        /// Просмотр объекта на карте
        /// </summary>
        /// <returns></returns>
        public Bitmap ShowOnMap()
        {
            Shaking();
            foreach (var ship in _setShips.GetShips())
            {
                return _map.CreateMap(_pictureWidth, _pictureHeight, ship);
            } 
            return new(_pictureWidth, _pictureHeight);
        }
        public string GetData(char separatorType, char separatorData)
        {
            string data = $"{_map.GetType().Name}{separatorType}";
            foreach (var ship in _setShips.GetShips())
            {
                data += $"{ship.GetInfo()}{separatorData}";
            }
            return data;
        }
        /// <summary>
        /// Загрузка списка из массива строк
        /// </summary>
        /// <param name="records"></param>
        public void LoadData(string[] records)
        {
            foreach (var rec in records)
            {
                _setShips.Insert(DrawingObjectShip.Create(rec) as T);
            }
        }
        /// <summary>
        /// Перемещение объекта по крате
        /// </summary>
        /// <param name="direction"></param>
        /// <returns></returns>
        public Bitmap MoveObject(Direction direction)
        {
            if (_map != null)
            {
                return _map.MoveObject(direction);
            }
            return new(_pictureWidth, _pictureHeight);
        }
        /// <summary>
        /// "Взбалтываем" набор, чтобы все элементы оказались в начале
        /// </summary>
        private void Shaking()
        {
            int j = _setShips.Count - 1;
            for (int i = 0; i < _setShips.Count; i++)
            {
                if (_setShips[i] == null)
                {
                    for (; j > i; j--)
                    {
                        var ship = _setShips[j];
                        if (ship != null)
                        {
                            _setShips.Insert(ship, i);
                            _setShips.Remove(j);
                            break;
                        }
                    }
                    if (j <= i)
                    {
                        return;
                    }
                }
            }
        }
        /// <summary>
        /// Метод отрисовки фона
        /// </summary>
        /// <param name="g"></param>
        private void DrawBackground(Graphics g)
        {
            Pen pen = new(Color.Brown, 3);
            g.FillRectangle(Brushes.Aqua, 0, 0, _pictureWidth, _pictureHeight);
            for (int i = 0; i < _pictureWidth / _placeSizeWidth; i++)
            {
                for (int j = 0; j < _pictureHeight / _placeSizeHeight + 1; ++j)
                {//линия рамзетки места
                    g.DrawLine(pen, i * _placeSizeWidth, j * _placeSizeHeight, i * _placeSizeWidth + _placeSizeWidth / 2, j * _placeSizeHeight);
                }
                g.DrawLine(pen, i * _placeSizeWidth, 0, i * _placeSizeWidth, (_pictureHeight / _placeSizeHeight) * _placeSizeHeight);
            }
        }
        /// <summary>
        /// Метод прорисовки объектов
        /// </summary>
        /// <param name="g"></param>
        private void DrawShips(Graphics g)
        {
            int i = 0;
            int countOfShipsInLine = _pictureWidth / _placeSizeWidth;
            int countOfShipsInColumn = _pictureHeight / _placeSizeHeight;
            foreach (var ship in _setShips.GetShips())
            {
                ship.SetObject((countOfShipsInLine - (i % countOfShipsInLine) - 1) * _placeSizeWidth, (countOfShipsInColumn - (i / countOfShipsInLine) - 1) * _placeSizeHeight, _pictureWidth, _pictureHeight);
                ship.DrawingObject(g);
                i++;
            }
        }
        public void Sort(IComparer<T> comparer)
        {
            _setShips.SortSet(comparer);
        }
        /// <summary>
        /// Перегрузка операторов
        /// </summary>
        /// <param name="map"></param>
        /// <param name="ship"></param>
        /// <returns></returns>
        public static int operator +(MapWithSetShipsGeneric<T, U> map, T ship)
        {
            return map._setShips.Insert(ship);
        }
        /// <summary>
        /// Перегрузка оператора вычитания
        /// </summary>
        /// <param name="map"></param>
        /// <param name="position"></param>
        /// <returns></returns>
        public static T operator -(MapWithSetShipsGeneric<T, U> map, int position)
        {
            return map._setShips.Remove(position);
        }
        public int add(T ship)
        {
            return _setShips.Insert(ship);
        }

    }
}
